#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for the 1D fixed bed module

Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    TransformationFactory,
    value,
    units as pyunits,
    Constraint,
    Var,
)
from pyomo.dae import ContinuousSet, Integral
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigBlock
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
import pyomo.common.unittest as unittest

import idaes
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.initialization import initialize_by_time_element
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.util.performance import PerformanceBaseClass

# Import FixedBed1D unit model
from idaes.models_extra.gas_solid_contactors.unit_models.fixed_bed_1D import FixedBed1D

# Import property packages
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.hetero_reactions import (
    HeteroReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 360], time_units=pyunits.s)

    # Set up thermo props and reaction props
    m.fs.gas_props = GasPhaseParameterBlock()
    m.fs.solid_props = SolidPhaseParameterBlock()
    m.fs.solid_rxns = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
    )

    m.fs.unit = FixedBed1D(
        gas_phase_config={"property_package": m.fs.gas_props},
        solid_phase_config={
            "property_package": m.fs.solid_props,
            "reaction_package": m.fs.solid_rxns,
        },
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 15

    assert m.fs.unit.config.dynamic is True
    assert m.fs.unit.config.has_holdup is True
    assert isinstance(m.fs.unit.config.gas_phase_config, ConfigBlock)
    assert isinstance(m.fs.unit.config.solid_phase_config, ConfigBlock)

    assert m.fs.unit.config.finite_elements == 10
    assert m.fs.unit.config.length_domain_set == [0.0, 1.0]
    assert m.fs.unit.config.transformation_method == "dae.finite_difference"
    assert m.fs.unit.config.transformation_scheme == "BACKWARD"
    assert m.fs.unit.config.collocation_points == 3
    assert m.fs.unit.config.flow_type == "forward_flow"
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.componentTotal
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.enthalpyTotal
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_pressure_change is True

    # Check gas phase config arguments
    assert len(m.fs.unit.config.gas_phase_config) == 7
    assert m.fs.unit.config.gas_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.gas_phase_config.property_package is m.fs.gas_props
    assert bool(m.fs.unit.config.gas_phase_config.property_package_args) is False
    assert m.fs.unit.config.gas_phase_config.reaction_package is None

    # Check solid phase config arguments
    assert len(m.fs.unit.config.solid_phase_config) == 7
    assert m.fs.unit.config.solid_phase_config.has_equilibrium_reactions is False
    assert m.fs.unit.config.solid_phase_config.property_package is m.fs.solid_props
    assert bool(m.fs.unit.config.solid_phase_config.property_package_args) is False
    assert m.fs.unit.config.solid_phase_config.reaction_package is m.fs.solid_rxns


@pytest.mark.unit
def test_config_validation():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 360], time_units=pyunits.s)

    # Set up thermo props and reaction props
    m.fs.gas_props = GasPhaseParameterBlock()
    m.fs.solid_props = SolidPhaseParameterBlock()
    m.fs.solid_rxns = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
    )

    m.fs.unit = FixedBed1D(
        gas_phase_config={"property_package": m.fs.gas_props},
        solid_phase_config={
            "property_package": m.fs.solid_props,
            "reaction_package": m.fs.solid_rxns,
        },
    )

    with pytest.raises(ConfigurationError):
        m.fs.unit = FixedBed1D(
            flow_type="reverse_flow",
            transformation_method="dae.collocation",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

    with pytest.raises(ConfigurationError):
        m.fs.unit = FixedBed1D(
            flow_type="reverse_flow",
            transformation_method="dae.finite_difference",
            transformation_scheme="BACKWARD",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

    with pytest.raises(ConfigurationError):
        m.fs.unit = FixedBed1D(
            flow_type="forward_flow",
            transformation_method="dae.finite_difference",
            transformation_scheme="FORWARD",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

    with pytest.raises(ConfigurationError):
        m.fs.unit = FixedBed1D(
            transformation_method="dae.collocation",
            transformation_scheme="BACKWARD",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

    with pytest.raises(ConfigurationError):
        m.fs.unit = FixedBed1D(
            transformation_method="dae.collocation",
            transformation_scheme="FORWARD",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )


# -----------------------------------------------------------------------------
def build_model():
    m = ConcreteModel()
    horizon = 360
    m.fs = FlowsheetBlock(dynamic=True, time_set=[0, horizon], time_units=pyunits.s)

    m.fs.gas_props = GasPhaseParameterBlock()
    m.fs.solid_props = SolidPhaseParameterBlock()
    m.fs.solid_rxns = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
    )

    m.fs.unit = FixedBed1D(
        gas_phase_config={"property_package": m.fs.gas_props},
        solid_phase_config={
            "property_package": m.fs.solid_props,
            "reaction_package": m.fs.solid_rxns,
        },
    )

    # Discretize time domain
    t_element_size = 120  # s
    ntfe = int(horizon / t_element_size)
    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

    # Set reactor design conditions
    m.fs.unit.bed_diameter.fix(9)  # diameter of the reactor [m]
    m.fs.unit.bed_height.fix(5)  # height of solids in the reactor [m]

    # Fix inlet port variables (boundary conditions) for gas
    for t in m.fs.time:
        m.fs.unit.gas_inlet.flow_mol[t].fix(20)  # mol/s
        m.fs.unit.gas_inlet.temperature[t].fix(723.15)  # K
        m.fs.unit.gas_inlet.pressure[t].fix(1.565e5)  # Pa
        m.fs.unit.gas_inlet.mole_frac_comp[t, "CO2"].fix(0.02499)
        m.fs.unit.gas_inlet.mole_frac_comp[t, "H2O"].fix(0.00001)
        m.fs.unit.gas_inlet.mole_frac_comp[t, "CH4"].fix(0.975)

    # Specify gas phase initial conditions
    t0 = m.fs.time.first()
    for x in m.fs.unit.length_domain:
        if x != m.fs.unit.length_domain.first():  # skip inlet as it's already fixed
            m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CO2"].fix(0.02499)
            m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["H2O"].fix(0.00001)
            m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CH4"].fix(0.975)
            m.fs.unit.gas_phase.properties[t0, x].temperature.fix(1273.15)  # K
            m.fs.unit.gas_phase.properties[t0, x].flow_mol.fix(20)  # mol/s

    # Specify solid phase initial conditions
    t0 = m.fs.time.first()
    for x in m.fs.unit.length_domain:
        m.fs.unit.solid_properties[t0, x].temperature.fix(1273.15)
        m.fs.unit.solid_properties[t0, x].particle_porosity.fix(0.20)
        m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe2O3"].fix(0.45)
        m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe3O4"].fix(0)
        m.fs.unit.solid_properties[t0, x].mass_frac_comp["Al2O3"].fix(0.55)

    return m


def set_scaling(iron_oc):
    # Set scaling gas phase for state variables
    iron_oc.fs.gas_props.set_default_scaling("flow_mol", 1e-3)
    iron_oc.fs.gas_props.set_default_scaling("pressure", 1e-5)
    iron_oc.fs.gas_props.set_default_scaling("temperature", 1e-2)
    for comp in iron_oc.fs.gas_props.component_list:
        iron_oc.fs.gas_props.set_default_scaling("mole_frac_comp", 1e1, index=comp)
    # Set scaling for gas phase thermophysical and transport properties
    iron_oc.fs.gas_props.set_default_scaling("enth_mol", 1e-6)
    iron_oc.fs.gas_props.set_default_scaling("enth_mol_comp", 1e-6)
    iron_oc.fs.gas_props.set_default_scaling("cp_mol", 1e-6)
    iron_oc.fs.gas_props.set_default_scaling("cp_mol_comp", 1e-6)
    iron_oc.fs.gas_props.set_default_scaling("cp_mass", 1e-6)
    iron_oc.fs.gas_props.set_default_scaling("entr_mol", 1e-2)
    iron_oc.fs.gas_props.set_default_scaling("entr_mol_phase", 1e-2)
    iron_oc.fs.gas_props.set_default_scaling("dens_mol", 1)
    iron_oc.fs.gas_props.set_default_scaling("dens_mol_comp", 1)
    iron_oc.fs.gas_props.set_default_scaling("dens_mass", 1e2)
    iron_oc.fs.gas_props.set_default_scaling("visc_d_comp", 1e4)
    iron_oc.fs.gas_props.set_default_scaling("diffusion_comp", 1e5)
    iron_oc.fs.gas_props.set_default_scaling("therm_cond_comp", 1e2)
    iron_oc.fs.gas_props.set_default_scaling("visc_d", 1e5)
    iron_oc.fs.gas_props.set_default_scaling("therm_cond", 1e0)
    iron_oc.fs.gas_props.set_default_scaling("mw", 1e2)

    # Set scaling for solid phase state variables
    iron_oc.fs.solid_props.set_default_scaling("particle_porosity", 1e2)
    iron_oc.fs.solid_props.set_default_scaling("temperature", 1e-2)
    for comp in iron_oc.fs.solid_props.component_list:
        iron_oc.fs.solid_props.set_default_scaling("mass_frac_comp", 1e1, index=comp)

    # Set scaling for solid phase thermophysical and transport properties
    iron_oc.fs.solid_props.set_default_scaling("enth_mass", 1e-6)
    iron_oc.fs.solid_props.set_default_scaling("enth_mol_comp", 1e-6)
    iron_oc.fs.solid_props.set_default_scaling("cp_mol_comp", 1e-6)
    iron_oc.fs.solid_props.set_default_scaling("cp_mass", 1e-6)
    iron_oc.fs.solid_props.set_default_scaling("dens_mass_particle", 1e-2)
    iron_oc.fs.solid_props.set_default_scaling("dens_mass_skeletal", 1e-2)

    FB1D = iron_oc.fs.unit  # alias to keep test lines short

    # Calculate scaling factors
    iscale.calculate_scaling_factors(FB1D)


@pytest.mark.performance
class Test_FixedBed1D_Performance(PerformanceBaseClass, unittest.TestCase):
    # TODO: Remove this once Pyomo bug in DAE units is fixed
    TEST_UNITS = False

    def build_model(self):
        model = build_model()
        set_scaling(model)
        return model

    def initialize_model(self, model):
        with idaes.temporary_config_ctx():
            optarg = {
                "tol": 1e-5,
                "bound_push": 1e-22,
                "nlp_scaling_method": "user-scaling",
            }
            model.fs.unit.initialize(optarg=optarg)

    def solve_model(self, model):
        with idaes.temporary_config_ctx():
            solver.options = {
                "tol": 1e-5,
                "bound_push": 1e-22,
                "nlp_scaling_method": "user-scaling",
            }
            results = solver.solve(model)

            # Check for optimal solution
            assert_optimal_termination(results)


class TestIronOC(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        return build_model()

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert isinstance(iron_oc.fs.unit.bed_area_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_area_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_area_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_super_vel, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_config_pressure_drop, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_comp_hetero_rxn, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_material_holdup_calculation, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_material_balances, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_sum_component_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_energy_holdup_calculation, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_enthalpy_balances, Constraint)

        assert isinstance(iron_oc.fs.unit.reynolds_number_particle, Constraint)
        assert isinstance(iron_oc.fs.unit.prandtl_number, Constraint)
        assert isinstance(iron_oc.fs.unit.nusselt_number_particle, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_solid_htc_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_heat_transfer, Constraint)

        assert number_variables(iron_oc) == 3402
        assert number_total_constraints(iron_oc) == 3202
        assert number_unused_variables(iron_oc) == 71

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.mark.component
    def test_scaling(self, iron_oc):
        set_scaling(iron_oc)

        FB1D = iron_oc.fs.unit  # alias to keep test lines short

        assert pytest.approx(0.01111, rel=1e-3) == iscale.get_scaling_factor(
            FB1D.bed_diameter
        )
        assert pytest.approx(1, rel=1e-3) == iscale.get_scaling_factor(FB1D.bed_height)

        for c in FB1D.bed_area_eqn.values():
            assert pytest.approx(
                0.0015719, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB1D.solid_phase_area_constraint.items():
            assert pytest.approx(
                0.0015719, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in FB1D.solid_material_holdup_calculation.items():
            assert pytest.approx(
                1.5719e-05, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in FB1D.solid_material_accumulation_disc_eq.items():
            assert pytest.approx(
                1e-07, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x, j), c in FB1D.solid_material_balances.items():
            print(j)
            sf = {
                "Fe2O3": 9.84345e-05,
                "Fe3O4": 6.78910e-05,
                "Al2O3": 0.000154168,
            }
            assert pytest.approx(
                sf[j], rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in FB1D.solid_sum_component_eqn.items():
            assert pytest.approx(
                10.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in FB1D.solid_energy_holdup_calculation.items():
            assert pytest.approx(
                1.5719e-11, rel=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, x), c in FB1D.solid_energy_accumulation_disc_eq.items():
            assert pytest.approx(
                1e-11, rel=1e-8
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        with idaes.temporary_config_ctx():
            optarg = {
                "tol": 1e-5,
                "bound_push": 1e-22,
                "nlp_scaling_method": "user-scaling",
            }

            initialization_tester(iron_oc, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_block_triangularization_initialization(self, iron_oc):
        iron_oc.fs.unit.block_triangularization_initialize(calc_var_kwds={"eps": 1e-5})

        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time(self, iron_oc):
        with idaes.temporary_config_ctx():
            optarg = {
                "tol": 1e-5,
                "bound_push": 1e-22,
                "nlp_scaling_method": "user-scaling",
            }
            solver = get_solver("ipopt", optarg)  # create solver

            initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        with idaes.temporary_config_ctx():
            solver.options = {
                "tol": 1e-5,
                "bound_push": 1e-22,
                "nlp_scaling_method": "user-scaling",
            }
            results = solver.solve(iron_oc)

            # Check for optimal solution
            assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        tf = iron_oc.fs.time.last()
        xf = iron_oc.fs.unit.length_domain.last()
        assert (
            pytest.approx(0.01209, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[t0, x0].value
        )
        assert (
            pytest.approx(0.06714, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[tf, xf].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(146126.06009, abs=1e-2)
            == iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(10373.9409, abs=1e-2)
            == iron_oc.fs.unit.gas_inlet.pressure[tf].value
            - iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(0.19996, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].particle_porosity.value
        )
        assert (
            pytest.approx(0.55007, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Al2O3"].value
        )
        assert (
            pytest.approx(0.44599, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe2O3"].value
        )
        assert (
            pytest.approx(0.00393951, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe3O4"].value
        )
        assert (
            pytest.approx(1272.86689, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].temperature.value
        )

    @pytest.mark.component
    @pytest.mark.xfail(reason="known DAE unit consistency issue, see Pyomo/pyomo#1790")
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()
        # no solid inlet/outlet ports (no stream table), test block reports instead
        t0 = iron_oc.fs.time.first()
        xf = iron_oc.fs.unit.length_domain.last()
        iron_oc.fs.unit.solid_properties[t0, xf].report()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iron_oc):
        perf_dict = iron_oc.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Bed Height": iron_oc.fs.unit.bed_height,
                "Bed Area": iron_oc.fs.unit.bed_area,
                "Gas Inlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 0],
                "Gas Outlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 1],
            }
        }

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        tf = iron_oc.fs.time.last()
        iron_oc.fs.unit.gas_outlet.flow_mol[tf].fix(0)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_reverse_flow(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        horizon = 360
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, horizon], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed1D(
            flow_type="reverse_flow",
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

        # Discretize time domain
        t_element_size = 120  # s
        ntfe = int(horizon / t_element_size)
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(9)  # diameter of the reactor [m]
        m.fs.unit.bed_height.fix(5)  # height of solids in the reactor [m]

        # Fix inlet port variables (boundary conditions) for gas
        for t in m.fs.time:
            m.fs.unit.gas_inlet.flow_mol[t].fix(20)  # mol/s
            m.fs.unit.gas_inlet.temperature[t].fix(723.15)  # K
            m.fs.unit.gas_inlet.pressure[t].fix(1.565e5)  # Pa
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CO2"].fix(0.02499)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "H2O"].fix(0.00001)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CH4"].fix(0.975)

        # Specify gas phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            if x != m.fs.unit.length_domain.last():  # skip inlet as it's already fixed
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CO2"].fix(0.02499)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["H2O"].fix(0.00001)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CH4"].fix(0.975)
                m.fs.unit.gas_phase.properties[t0, x].temperature.fix(1273.15)  # K
                m.fs.unit.gas_phase.properties[t0, x].flow_mol.fix(20)  # mol/s

        # Specify solid phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            m.fs.unit.solid_properties[t0, x].temperature.fix(1273.15)
            m.fs.unit.solid_properties[t0, x].particle_porosity.fix(0.20)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe2O3"].fix(0.45)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe3O4"].fix(0)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Al2O3"].fix(0.55)

        return m

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        optarg = {
            "tol": 1e-5,
            "bound_push": 1e-22,
            "nlp_scaling_method": "user-scaling",
        }

        initialization_tester(iron_oc, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_block_triangularization_initialization(self, iron_oc):
        iron_oc.fs.unit.block_triangularization_initialize(calc_var_kwds={"eps": 1e-5})

        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time(self, iron_oc):

        optarg = {
            "tol": 1e-5,
            "bound_push": 1e-22,
            "nlp_scaling_method": "user-scaling",
        }
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        tf = iron_oc.fs.time.last()
        xf = iron_oc.fs.unit.length_domain.last()
        assert (
            pytest.approx(0.01209, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[t0, x0].value
        )
        assert (
            pytest.approx(0.012078, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[tf, xf].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(146126.06009, abs=1e-2)
            == iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(10373.9409, abs=1e-2)
            == iron_oc.fs.unit.gas_inlet.pressure[tf].value
            - iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(0.196012, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].particle_porosity.value
        )
        assert (
            pytest.approx(0.55730, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Al2O3"].value
        )
        assert (
            pytest.approx(0.05858366, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe2O3"].value
        )
        assert (
            pytest.approx(0.38411633, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe3O4"].value
        )
        assert (
            pytest.approx(723.266955, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].temperature.value
        )

    @pytest.mark.component
    @pytest.mark.xfail(reason="known DAE unit consistency issue, see Pyomo/pyomo#1790")
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()
        # no solid inlet/outlet ports (no stream table), test block reports instead
        t0 = iron_oc.fs.time.first()
        xf = iron_oc.fs.unit.length_domain.last()
        iron_oc.fs.unit.solid_properties[t0, xf].report()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iron_oc):
        perf_dict = iron_oc.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Bed Height": iron_oc.fs.unit.bed_height,
                "Bed Area": iron_oc.fs.unit.bed_area,
                "Gas Inlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 0],
                "Gas Outlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 1],
            }
        }

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        tf = iron_oc.fs.time.last()
        iron_oc.fs.unit.gas_outlet.flow_mol[tf].fix(0)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


# -----------------------------------------------------------------------------
class TestIronOC_NoReaction(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        horizon = 360
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, horizon], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()

        m.fs.unit = FixedBed1D(
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={"property_package": m.fs.solid_props},
        )

        # Discretize time domain
        t_element_size = 120  # s
        ntfe = int(horizon / t_element_size)
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(9)  # diameter of the reactor [m]
        m.fs.unit.bed_height.fix(5)  # height of solids in the reactor [m]

        # Fix inlet port variables (boundary conditions) for gas
        for t in m.fs.time:
            m.fs.unit.gas_inlet.flow_mol[t].fix(20)  # mol/s
            m.fs.unit.gas_inlet.temperature[t].fix(1273.15)  # K
            m.fs.unit.gas_inlet.pressure[t].fix(1.565e5)  # Pa
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CO2"].fix(0.02499)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "H2O"].fix(0.00001)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CH4"].fix(0.975)

        # Specify gas phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            if x != m.fs.unit.length_domain.first():  # skip inlet as it's already fixed
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CO2"].fix(0.02499)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["H2O"].fix(0.00001)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CH4"].fix(0.975)
                m.fs.unit.gas_phase.properties[t0, x].temperature.fix(1273.15)  # K
                m.fs.unit.gas_phase.properties[t0, x].flow_mol.fix(20)  # mol/s

        # Specify solid phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            m.fs.unit.solid_properties[t0, x].temperature.fix(1273.15)
            m.fs.unit.solid_properties[t0, x].particle_porosity.fix(0.20)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe2O3"].fix(0.45)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe3O4"].fix(0)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Al2O3"].fix(0.55)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert hasattr(iron_oc.fs.unit, "gas_inlet")
        assert len(iron_oc.fs.unit.gas_inlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_inlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_inlet.pressure, Var)

        assert hasattr(iron_oc.fs.unit, "gas_outlet")
        assert len(iron_oc.fs.unit.gas_outlet.vars) == 4
        assert isinstance(iron_oc.fs.unit.gas_outlet.flow_mol, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.mole_frac_comp, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.temperature, Var)
        assert isinstance(iron_oc.fs.unit.gas_outlet.pressure, Var)

        assert isinstance(iron_oc.fs.unit.bed_area_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_area_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_area_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_super_vel, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_config_pressure_drop, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_material_holdup_calculation, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_material_balances, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_sum_component_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_phase_heat_transfer, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_energy_holdup_calculation, Constraint)
        assert isinstance(iron_oc.fs.unit.solid_enthalpy_balances, Constraint)

        assert isinstance(iron_oc.fs.unit.reynolds_number_particle, Constraint)
        assert isinstance(iron_oc.fs.unit.prandtl_number, Constraint)
        assert isinstance(iron_oc.fs.unit.nusselt_number_particle, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_solid_htc_eqn, Constraint)
        assert isinstance(iron_oc.fs.unit.gas_phase_heat_transfer, Constraint)

        assert number_variables(iron_oc) == 3088
        assert number_total_constraints(iron_oc) == 2894
        assert number_unused_variables(iron_oc) == 71

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        optarg = {
            "tol": 1e-5,
            "bound_push": 1e-22,
            "nlp_scaling_method": "user-scaling",
        }
        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        initialization_tester(iron_oc, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_block_triangularization_initialization(self, iron_oc):
        iron_oc.fs.unit.block_triangularization_initialize(calc_var_kwds={"eps": 1e-5})

        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time(self, iron_oc):

        optarg = {
            "tol": 1e-5,
            "bound_push": 1e-22,
            "nlp_scaling_method": "user-scaling",
        }
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        assert iron_oc.fs.unit.solid_properties[t0, x0].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solid_properties[
                    t, x0
                ].particle_porosity.fixed
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_particle_constraint.active
            assert iron_oc.fs.unit.solid_properties[
                t, x0
            ].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        t0 = iron_oc.fs.time.first()
        x0 = iron_oc.fs.unit.length_domain.first()
        tf = iron_oc.fs.time.last()
        xf = iron_oc.fs.unit.length_domain.last()
        assert (
            pytest.approx(0.01209, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[t0, x0].value
        )
        assert (
            pytest.approx(0.02159, abs=1e-2)
            == iron_oc.fs.unit.velocity_superficial_gas[tf, xf].value
        )
        # Check the pressure drop that occurs across the bed
        assert (
            pytest.approx(154128.50977, abs=1e-2)
            == iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(2371.49023, abs=1e-2)
            == iron_oc.fs.unit.gas_inlet.pressure[tf].value
            - iron_oc.fs.unit.gas_outlet.pressure[tf].value
        )
        assert (
            pytest.approx(0.2, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].particle_porosity.value
        )
        assert (
            pytest.approx(0.55, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Al2O3"].value
        )
        assert (
            pytest.approx(0.45, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe2O3"].value
        )
        assert (
            pytest.approx(0.00, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].mass_frac_comp["Fe3O4"].value
        )
        assert (
            pytest.approx(1273.15, rel=1e-5)
            == iron_oc.fs.unit.solid_properties[tf, xf].temperature.value
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):

        iron_oc.fs.unit.time_set = ContinuousSet(
            bounds=(iron_oc.fs.time.first(), iron_oc.fs.time.last()),
            initialize=iron_oc.fs.time.get_finite_elements(),
            doc="time domain",
        )

        ###########################################################################
        #  Mass conservation check
        ###########################################################################
        def gas_input_total(b, t):
            calculate_variable_from_constraint(
                b.gas_phase.properties[t, 0].mw,
                b.gas_phase.properties[t, 0].mw_eqn,
            )
            return b.gas_phase.properties[t, 0].mw * b.gas_inlet.flow_mol[t]

        iron_oc.fs.unit.gas_input_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_input_total,
            doc="Gas input total, kg",
        )

        def gas_output_total(b, t):
            calculate_variable_from_constraint(
                b.gas_phase.properties[t, 1].mw,
                b.gas_phase.properties[t, 1].mw_eqn,
            )
            return b.gas_phase.properties[t, 1].mw * b.gas_outlet.flow_mol[t]

        iron_oc.fs.unit.gas_output_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_output_total,
            doc="Gas output total, kg",
        )

        def gas_holdup(b, x):
            tf = b.flowsheet().time.last()
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.gas_phase.properties[tf, x].mw,
                b.gas_phase.properties[tf, x].mw_eqn,
            )
            calculate_variable_from_constraint(
                b.gas_phase.properties[t0, x].mw,
                b.gas_phase.properties[t0, x].mw_eqn,
            )
            mass_final_per_length = (
                b.gas_phase.properties[tf, x].dens_mol
                * b.gas_phase.properties[tf, 1].mw
                * b.gas_phase.area[tf, x]
            )
            mass_initial_per_length = (
                b.gas_phase.properties[t0, x].dens_mol
                * b.gas_phase.properties[t0, 1].mw
                * b.gas_phase.area[t0, x]
            )
            return (mass_final_per_length - mass_initial_per_length) * b.bed_height

        iron_oc.fs.unit.gas_holdup = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=gas_holdup,
            doc="Gas mass holdup, kg",
        )

        def oc_at_initial_time(b, x):
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.solid_properties[t0, x].dens_mass_particle,
                b.solid_properties[t0, x].density_particle_constraint,
            )
            mass_per_length = (
                b.solid_properties[t0, x].dens_mass_particle * b.solid_phase_area[t0, x]
            )
            return mass_per_length * b.bed_height

        iron_oc.fs.unit.oc_at_initial_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_at_initial_time,
            doc="OC at initial reaction time, kg",
        )

        def oc_at_final_time(b, x):
            tf = b.flowsheet().time.last()
            calculate_variable_from_constraint(
                b.solid_properties[tf, x].dens_mass_particle,
                b.solid_properties[tf, x].density_particle_constraint,
            )
            mass_per_length = (
                b.solid_properties[tf, x].dens_mass_particle * b.solid_phase_area[tf, x]
            )
            return mass_per_length * b.bed_height

        iron_oc.fs.unit.oc_at_final_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_at_final_time,
            doc="OC at final reaction time, kg",
        )

        mass_change_gas = value(
            iron_oc.fs.unit.gas_input_total
            - iron_oc.fs.unit.gas_output_total
            - iron_oc.fs.unit.gas_holdup
        )
        mass_change_solid = value(
            iron_oc.fs.unit.oc_at_initial_time - iron_oc.fs.unit.oc_at_final_time
        )

        assert abs(mass_change_gas) <= 1e-2
        assert abs(mass_change_solid) <= 1e-2

        ###########################################################################
        #  Energy conservation check
        ###########################################################################
        def ch4_reacted_mol(b, t):
            return (
                b.gas_inlet.flow_mol[t] * b.gas_inlet.mole_frac_comp[t, "CH4"]
                - b.gas_outlet.flow_mol[t] * b.gas_outlet.mole_frac_comp[t, "CH4"]
            )

        iron_oc.fs.unit.ch4_reacted_mol = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=ch4_reacted_mol,
            doc="CH4 reacted total, mol",
        )

        def fe2o3_at_initial_time_mol(b, x):
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.solid_properties[t0, x].dens_mass_particle,
                b.solid_properties[t0, x].density_particle_constraint,
            )
            mol_fe2o3_per_length = (
                b.solid_properties[t0, x].dens_mass_particle
                * b.solid_properties[t0, x].mass_frac_comp["Fe2O3"]
                * b.solid_phase_area[t0, x]
                / b.solid_properties[t0, x]._params.mw_comp["Fe2O3"]
            )
            return mol_fe2o3_per_length * b.bed_height

        iron_oc.fs.unit.fe2o3_at_initial_time_mol = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=fe2o3_at_initial_time_mol,
            doc="Fe2O3 at initial reaction time, mol",
        )

        def fe2o3_at_final_time_mol(b, x):
            tf = b.flowsheet().time.last()
            calculate_variable_from_constraint(
                b.solid_properties[tf, x].dens_mass_particle,
                b.solid_properties[tf, x].density_particle_constraint,
            )
            mol_fe2o3_per_length = (
                b.solid_properties[tf, x].dens_mass_particle
                * b.solid_properties[tf, x].mass_frac_comp["Fe2O3"]
                * b.solid_phase_area[tf, x]
                / b.solid_properties[tf, x]._params.mw_comp["Fe2O3"]
            )
            return mol_fe2o3_per_length * b.bed_height

        iron_oc.fs.unit.fe2o3_at_final_time_mol = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=fe2o3_at_final_time_mol,
            doc="Fe2O3 at final reaction time, mol",
        )

        def gas_input_energy_total(b, t):
            return b.gas_phase.properties[t, 0].enth_mol * b.gas_inlet.flow_mol[t]

        iron_oc.fs.unit.gas_input_energy_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_input_energy_total,
            doc="Gas input energy total, J",
        )

        def gas_output_energy_total(b, t):
            return b.gas_phase.properties[t, 1].enth_mol * b.gas_outlet.flow_mol[t]

        iron_oc.fs.unit.gas_output_energy_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_output_energy_total,
            doc="Gas output energy total, J",
        )

        def gas_energy_holdup(b, x):
            tf = b.flowsheet().time.last()
            t0 = b.flowsheet().time.first()
            energy_final_per_length = (
                b.gas_phase.properties[tf, x].dens_mol
                * b.gas_phase.properties[tf, x].enth_mol
                * b.gas_phase.area[tf, x]
            )
            energy_initial_per_length = (
                b.gas_phase.properties[t0, x].dens_mol
                * b.gas_phase.properties[t0, x].enth_mol
                * b.gas_phase.area[t0, x]
            )
            return (energy_final_per_length - energy_initial_per_length) * b.bed_height

        iron_oc.fs.unit.gas_energy_holdup = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=gas_energy_holdup,
            doc="Gas energy holdup, J",
        )

        def oc_energy_at_initial_time(b, x):
            t0 = b.flowsheet().time.first()
            energy_per_length = (
                b.solid_properties[t0, x].dens_mass_particle
                * b.solid_properties[t0, x].enth_mass
                * b.solid_phase_area[t0, x]
            )
            return energy_per_length * b.bed_height

        iron_oc.fs.unit.oc_energy_at_initial_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_energy_at_initial_time,
            doc="OC energy at initial reaction time, J",
        )

        def oc_energy_at_final_time(b, x):
            tf = b.flowsheet().time.last()
            energy_per_length = (
                b.solid_properties[tf, x].dens_mass_particle
                * b.solid_properties[tf, x].enth_mass
                * b.solid_phase_area[tf, x]
            )
            return energy_per_length * b.bed_height

        iron_oc.fs.unit.oc_energy_at_final_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_energy_at_final_time,
            doc="OC at final reaction time, J",
        )

        if (
            iron_oc.fs.unit.config.energy_balance_type is not EnergyBalanceType.none
            and iron_oc.fs.unit.config.solid_phase_config.reaction_package is not None
        ):
            e_reaction = (
                value(
                    (
                        iron_oc.fs.unit.fe2o3_at_final_time_mol
                        - iron_oc.fs.unit.fe2o3_at_initial_time_mol
                    )
                    * iron_oc.fs.unit.solid_reactions[0, 1]._params.dh_rxn["R1"]
                )
                / 12
            )
        else:
            e_reaction = 0

        enthalpy_change_gas = value(  # in - out - holdup/accumulation
            iron_oc.fs.unit.gas_input_energy_total
            - iron_oc.fs.unit.gas_output_energy_total
            - iron_oc.fs.unit.gas_energy_holdup
        )

        enthalpy_change_solid = value(
            iron_oc.fs.unit.oc_energy_at_final_time
            - iron_oc.fs.unit.oc_energy_at_initial_time
            - e_reaction
        )

        ebal_abs_tol = enthalpy_change_solid - enthalpy_change_gas

        assert abs(ebal_abs_tol) <= 1e-8

    @pytest.mark.component
    @pytest.mark.xfail(reason="known DAE unit consistency issue, see Pyomo/pyomo#1790")
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        iron_oc.fs.unit.report()
        # no solid inlet/outlet ports (no stream table), test block reports instead
        t0 = iron_oc.fs.time.first()
        xf = iron_oc.fs.unit.length_domain.last()
        iron_oc.fs.unit.solid_properties[t0, xf].report()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iron_oc):
        perf_dict = iron_oc.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Bed Height": iron_oc.fs.unit.bed_height,
                "Bed Area": iron_oc.fs.unit.bed_area,
                "Gas Inlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 0],
                "Gas Outlet Velocity": iron_oc.fs.unit.velocity_superficial_gas[0, 1],
            }
        }

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        tf = iron_oc.fs.time.last()
        iron_oc.fs.unit.gas_outlet.flow_mol[tf].fix(0)

        with pytest.raises(InitializationError):
            iron_oc.fs.unit.initialize()


class TestIronOC_conservation_with_reaction(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        horizon = 3600
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, horizon], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed1D(
            finite_elements=20,
            gas_phase_config={"property_package": m.fs.gas_props},
            solid_phase_config={
                "property_package": m.fs.solid_props,
                "reaction_package": m.fs.solid_rxns,
            },
        )

        # Discretize time domain
        t_element_size = 120  # s
        ntfe = int(horizon / t_element_size)
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(9)  # diameter of the reactor [m]
        m.fs.unit.bed_height.fix(5)  # height of solids in the reactor [m]

        # Fix inlet port variables (boundary conditions) for gas
        for t in m.fs.time:
            m.fs.unit.gas_inlet.flow_mol[t].fix(20)  # mol/s
            m.fs.unit.gas_inlet.temperature[t].fix(1273.15)  # K
            m.fs.unit.gas_inlet.pressure[t].fix(1.565e5)  # Pa
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CO2"].fix(0.02499)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "H2O"].fix(0.00001)
            m.fs.unit.gas_inlet.mole_frac_comp[t, "CH4"].fix(0.975)

        # Specify gas phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            if x != m.fs.unit.length_domain.first():  # skip inlet as it's already fixed
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CO2"].fix(0.02499)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["H2O"].fix(0.00001)
                m.fs.unit.gas_phase.properties[t0, x].mole_frac_comp["CH4"].fix(0.975)
                m.fs.unit.gas_phase.properties[t0, x].temperature.fix(1273.15)  # K
                m.fs.unit.gas_phase.properties[t0, x].flow_mol.fix(20)  # mol/s

        # Specify solid phase initial conditions
        t0 = m.fs.time.first()
        for x in m.fs.unit.length_domain:
            m.fs.unit.solid_properties[t0, x].temperature.fix(1273.15)
            m.fs.unit.solid_properties[t0, x].particle_porosity.fix(0.20)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe2O3"].fix(0.45)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Fe3O4"].fix(0)
            m.fs.unit.solid_properties[t0, x].mass_frac_comp["Al2O3"].fix(0.55)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.integration
    def test_conservation(self, iron_oc):
        # Calculate scaling factors
        iscale.calculate_scaling_factors(iron_oc)

        # Initialize model
        iron_oc.fs.unit.block_triangularization_initialize(calc_var_kwds={"eps": 1e-5})

        # Solve model
        initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        iron_oc.fs.unit.time_set = ContinuousSet(
            bounds=(iron_oc.fs.time.first(), iron_oc.fs.time.last()),
            initialize=iron_oc.fs.time.get_finite_elements(),
            doc="time domain",
        )

        ###########################################################################
        #  Mass conservation check
        ###########################################################################
        def gas_input_total(b, t):
            calculate_variable_from_constraint(
                b.gas_phase.properties[t, 0].mw,
                b.gas_phase.properties[t, 0].mw_eqn,
            )
            return b.gas_phase.properties[t, 0].mw * b.gas_inlet.flow_mol[t]

        iron_oc.fs.unit.gas_input_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_input_total,
            doc="Gas input total, kg",
        )

        def gas_output_total(b, t):
            calculate_variable_from_constraint(
                b.gas_phase.properties[t, 1].mw,
                b.gas_phase.properties[t, 1].mw_eqn,
            )
            return b.gas_phase.properties[t, 1].mw * b.gas_outlet.flow_mol[t]

        iron_oc.fs.unit.gas_output_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_output_total,
            doc="Gas output total, kg",
        )

        def gas_holdup(b, x):
            tf = b.flowsheet().time.last()
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.gas_phase.properties[tf, x].mw,
                b.gas_phase.properties[tf, x].mw_eqn,
            )
            calculate_variable_from_constraint(
                b.gas_phase.properties[t0, x].mw,
                b.gas_phase.properties[t0, x].mw_eqn,
            )
            mass_final_per_length = (
                b.gas_phase.properties[tf, x].dens_mol
                * b.gas_phase.properties[tf, 1].mw
                * b.gas_phase.area[tf, x]
            )
            mass_initial_per_length = (
                b.gas_phase.properties[t0, x].dens_mol
                * b.gas_phase.properties[t0, 1].mw
                * b.gas_phase.area[t0, x]
            )
            return (mass_final_per_length - mass_initial_per_length) * b.bed_height

        iron_oc.fs.unit.gas_holdup = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=gas_holdup,
            doc="Gas mass holdup, kg",
        )

        def oc_at_initial_time(b, x):
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.solid_properties[t0, x].dens_mass_particle,
                b.solid_properties[t0, x].density_particle_constraint,
            )
            mass_per_length = (
                b.solid_properties[t0, x].dens_mass_particle * b.solid_phase_area[t0, x]
            )
            return mass_per_length * b.bed_height

        iron_oc.fs.unit.oc_at_initial_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_at_initial_time,
            doc="OC at initial reaction time, kg",
        )

        def oc_at_final_time(b, x):
            tf = b.flowsheet().time.last()
            calculate_variable_from_constraint(
                b.solid_properties[tf, x].dens_mass_particle,
                b.solid_properties[tf, x].density_particle_constraint,
            )
            mass_per_length = (
                b.solid_properties[tf, x].dens_mass_particle * b.solid_phase_area[tf, x]
            )
            return mass_per_length * b.bed_height

        iron_oc.fs.unit.oc_at_final_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_at_final_time,
            doc="OC at final reaction time, kg",
        )

        mass_change_gas = value(
            iron_oc.fs.unit.gas_input_total
            - iron_oc.fs.unit.gas_output_total
            - iron_oc.fs.unit.gas_holdup
        )
        mass_change_solid = value(
            iron_oc.fs.unit.oc_at_initial_time - iron_oc.fs.unit.oc_at_final_time
        )
        mbal_abs_tol = abs(mass_change_gas + mass_change_solid)
        mbal_rel_tol = mbal_abs_tol / mass_change_solid
        assert abs(mbal_rel_tol) <= 0.15

        ###########################################################################
        #  Energy conservation check
        ###########################################################################
        def ch4_reacted_mol(b, t):
            return (
                b.gas_inlet.flow_mol[t] * b.gas_inlet.mole_frac_comp[t, "CH4"]
                - b.gas_outlet.flow_mol[t] * b.gas_outlet.mole_frac_comp[t, "CH4"]
            )

        iron_oc.fs.unit.ch4_reacted_mol = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=ch4_reacted_mol,
            doc="CH4 reacted total, mol",
        )

        def fe2o3_at_initial_time_mol(b, x):
            t0 = b.flowsheet().time.first()
            calculate_variable_from_constraint(
                b.solid_properties[t0, x].dens_mass_particle,
                b.solid_properties[t0, x].density_particle_constraint,
            )
            mol_fe2o3_per_length = (
                b.solid_properties[t0, x].dens_mass_particle
                * b.solid_properties[t0, x].mass_frac_comp["Fe2O3"]
                * b.solid_phase_area[t0, x]
                / b.solid_properties[t0, x]._params.mw_comp["Fe2O3"]
            )
            return mol_fe2o3_per_length * b.bed_height

        iron_oc.fs.unit.fe2o3_at_initial_time_mol = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=fe2o3_at_initial_time_mol,
            doc="Fe2O3 at initial reaction time, mol",
        )

        def fe2o3_at_final_time_mol(b, x):
            tf = b.flowsheet().time.last()
            calculate_variable_from_constraint(
                b.solid_properties[tf, x].dens_mass_particle,
                b.solid_properties[tf, x].density_particle_constraint,
            )
            mol_fe2o3_per_length = (
                b.solid_properties[tf, x].dens_mass_particle
                * b.solid_properties[tf, x].mass_frac_comp["Fe2O3"]
                * b.solid_phase_area[tf, x]
                / b.solid_properties[tf, x]._params.mw_comp["Fe2O3"]
            )
            return mol_fe2o3_per_length * b.bed_height

        iron_oc.fs.unit.fe2o3_at_final_time_mol = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=fe2o3_at_final_time_mol,
            doc="Fe2O3 at final reaction time, mol",
        )

        def gas_input_energy_total(b, t):
            return b.gas_phase.properties[t, 0].enth_mol * b.gas_inlet.flow_mol[t]

        iron_oc.fs.unit.gas_input_energy_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_input_energy_total,
            doc="Gas input energy total, J",
        )

        def gas_output_energy_total(b, t):
            return b.gas_phase.properties[t, 1].enth_mol * b.gas_outlet.flow_mol[t]

        iron_oc.fs.unit.gas_output_energy_total = Integral(
            iron_oc.fs.unit.time_set,
            wrt=iron_oc.fs.unit.time_set,
            rule=gas_output_energy_total,
            doc="Gas output energy total, J",
        )

        def gas_energy_holdup(b, x):
            tf = b.flowsheet().time.last()
            t0 = b.flowsheet().time.first()
            energy_final_per_length = (
                b.gas_phase.properties[tf, x].dens_mol
                * b.gas_phase.properties[tf, x].enth_mol
                * b.gas_phase.area[tf, x]
            )
            energy_initial_per_length = (
                b.gas_phase.properties[t0, x].dens_mol
                * b.gas_phase.properties[t0, x].enth_mol
                * b.gas_phase.area[t0, x]
            )
            return (energy_final_per_length - energy_initial_per_length) * b.bed_height

        iron_oc.fs.unit.gas_energy_holdup = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=gas_energy_holdup,
            doc="Gas energy holdup, J",
        )

        def oc_energy_at_initial_time(b, x):
            t0 = b.flowsheet().time.first()
            energy_per_length = (
                b.solid_properties[t0, x].dens_mass_particle
                * b.solid_properties[t0, x].enth_mass
                * b.solid_phase_area[t0, x]
            )
            return energy_per_length * b.bed_height

        iron_oc.fs.unit.oc_energy_at_initial_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_energy_at_initial_time,
            doc="OC energy at initial reaction time, J",
        )

        def oc_energy_at_final_time(b, x):
            tf = b.flowsheet().time.last()
            energy_per_length = (
                b.solid_properties[tf, x].dens_mass_particle
                * b.solid_properties[tf, x].enth_mass
                * b.solid_phase_area[tf, x]
            )
            return energy_per_length * b.bed_height

        iron_oc.fs.unit.oc_energy_at_final_time = Integral(
            iron_oc.fs.unit.length_domain,
            wrt=iron_oc.fs.unit.length_domain,
            rule=oc_energy_at_final_time,
            doc="OC at final reaction time, J",
        )

        if iron_oc.fs.unit.config.solid_phase_config.reaction_package is not None:
            e_reaction = (
                value(
                    (
                        iron_oc.fs.unit.fe2o3_at_final_time_mol
                        - iron_oc.fs.unit.fe2o3_at_initial_time_mol
                    )
                    * iron_oc.fs.unit.solid_reactions[0, 1]._params.dh_rxn["R1"]
                )
                / 12
            )
        else:
            e_reaction = 0

        enthalpy_change_gas = value(  # in - out - holdup/accumulation
            iron_oc.fs.unit.gas_input_energy_total
            - iron_oc.fs.unit.gas_output_energy_total
            - iron_oc.fs.unit.gas_energy_holdup
        )

        enthalpy_change_solid = value(
            iron_oc.fs.unit.oc_energy_at_final_time
            - iron_oc.fs.unit.oc_energy_at_initial_time
            - e_reaction
        )

        ebal_abs_tol = enthalpy_change_solid - enthalpy_change_gas
        ebal_rel_tol = ebal_abs_tol / enthalpy_change_solid
        assert abs(ebal_rel_tol) <= 0.15



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard plotting methods (scatter, parity and residual plots).

Author: B. Paul
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import statements
from itertools import combinations  # used to pick xvar pairs in scatter3D

import numpy as np
import matplotlib.pyplot as plt

# TODO: Consider adding seaborn in future
from matplotlib.backends.backend_pdf import PdfPages


def surrogate_scatter2D(surrogate, dataframe, filename=None, show=True):
    input_data = dataframe[surrogate.input_labels()]
    output_data = dataframe[surrogate.output_labels()]
    output_surrogate = surrogate.evaluate_surrogate(input_data)
    return _scatter2D(
        xdata=input_data.values,
        zdata=output_data.values,
        zfit=output_surrogate.values,
        xlabels=surrogate.input_labels(),
        zlabels=surrogate.output_labels(),
        show=show,
        filename=filename,
    )


def _scatter2D(
    xdata, zdata, zfit, xlabels=None, zlabels=None, show=True, filename=None
):
    """
    Plots training data as 2D scatter, with overlaid model output data.

    Required: input variable data, output variable data, model fit data

    Optional: input and output labels (will be generated as needed), filename
    to save plots to PDF

    Generally, x (input) variables are indexed by 'i', and z (output) variables
    are indexed by 'j'.
    """

    numins = np.shape(xdata)[1]  # number of input variables
    numouts = np.shape(zdata)[1]  # number of output variables

    # check and add labels if missing
    if xlabels is None:
        xlabels = []
        for i in range(numins):
            xlabels.append("x" + str(i + 1))
    if zlabels is None:
        zlabels = []
        for j in range(numouts):
            zlabels.append("z" + str(j + 1))

    if filename is not None:
        if filename[-4:] != ".pdf":  # checking if extension is present
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Filename does not end in .pdf, please amend "
                "filename with the proper extension."
            )
        pdf = PdfPages(filename)

    fig_list = []

    for j in range(numouts):  # loop over all outputs, zj
        for i in range(numins):  # plot every possible zj = f(xi)
            fig = plt.figure()
            ax = fig.add_subplot()
            ax.scatter(xdata[:, i], zdata[:, j], c="grey", marker="s", label="Data")
            ax.scatter(xdata[:, i], zfit[:, j], c="b", marker=".", label="Model")

            ax.set_xlabel(xlabels[i])
            ax.set_ylabel(zlabels[j])
            ax.set_title("2D Scatter Plot")
            ax.legend()

            if show is True:
                plt.show()
            if filename is not None:
                pdf.savefig(fig)

            fig_list.append(fig)

    if filename is not None:  # place outside loop to avoid closing/reopening
        pdf.close()

    return fig_list


def surrogate_scatter3D(surrogate, dataframe, filename=None, show=True):
    input_data = dataframe[surrogate.input_labels()]
    output_data = dataframe[surrogate.output_labels()]
    output_surrogate = surrogate.evaluate_surrogate(input_data)
    return _scatter3D(
        xdata=input_data.values,
        zdata=output_data.values,
        zfit=output_surrogate.values,
        xlabels=surrogate.input_labels(),
        zlabels=surrogate.output_labels(),
        show=show,
        filename=filename,
    )


def _scatter3D(
    xdata, zdata, zfit, xlabels=None, zlabels=None, show=True, filename=None
):
    """
    Plots training data as 3D scatter, with overlaid model output data.

    Required: input variable data, output variable data, model fit data

    Optional: input and output labels (will be generated as needed), filename
    to save plots to PDF

    Generally, x (input) variables are indexed by 'i', and z (output) variables
    are indexed by 'j'.
    """

    numins = np.shape(xdata)[1]  # number of input variables
    numouts = np.shape(zdata)[1]  # number of output variables

    # check and add labels if missing
    if xlabels is None:
        xlabels = []
        for i in range(numins):
            xlabels.append("x" + str(i + 1))
    if zlabels is None:
        zlabels = []
        for j in range(numouts):
            zlabels.append("z" + str(j + 1))

    if filename is not None:
        if filename[-4:] != ".pdf":  # checking if extension is present
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Filename does not end in .pdf, please amend "
                "filename with the proper extension."
            )
        pdf = PdfPages(filename)

    fig_list = []

    for j in range(numouts):  # loop over all outputs, zj
        for pair in list(combinations(range(numins), 2)):  # pick two x vars
            a, b = pair[0], pair[1]  # indices for the x variables picked
            fig = plt.figure()
            ax = fig.add_subplot(projection="3d")

            ax.scatter(
                xdata[:, a],
                xdata[:, b],
                zdata[:, j],
                c="grey",
                marker="s",
                label="Data",
            )
            ax.scatter(
                xdata[:, a], xdata[:, b], zfit[:, j], c="b", marker=".", label="Model"
            )
            ax.set_xlabel(xlabels[a])
            ax.set_ylabel(xlabels[b])
            ax.set_zlabel(zlabels[j])
            ax.set_title("3D Scatter Plot")
            ax.legend()

            if show is True:
                plt.show()
            if filename is not None:
                pdf.savefig(fig)

            fig_list.append(fig)

    if filename is not None:  # place outside loop to avoid closing/reopening
        pdf.close()

    return fig_list


def surrogate_parity(surrogate, dataframe, filename=None, show=True):
    input_data = dataframe[surrogate.input_labels()]
    output_data = dataframe[surrogate.output_labels()]
    output_surrogate = surrogate.evaluate_surrogate(input_data)
    return _parity(
        zdata=output_data.values,
        zfit=output_surrogate.values,
        zlabels=surrogate.output_labels(),
        show=show,
        filename=filename,
    )


def _parity(zdata, zfit, zlabels=None, show=True, filename=None):
    """
    Plots model output against data output.

    Required: output variable data, model fit data

    Optional: output labels (will be generated as needed), filename to save
    plots to PDF

    Generally, x (input) variables are indexed by 'i', and z (output) variables
    are indexed by 'j'.
    """
    numouts = np.shape(zdata)[1]  # number of output variables

    # check and add labels if missing

    if zlabels is None:
        zlabels = []
        for j in range(numouts):
            zlabels.append("z" + str(j + 1))

    if filename is not None:
        if filename[-4:] != ".pdf":  # checking if extension is present
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Filename does not end in .pdf, please amend "
                "filename with the proper extension."
            )
        pdf = PdfPages(filename)

    fig_list = []

    for j in range(numouts):  # loop over all outputs, zj
        fig = plt.figure()
        ax = fig.add_subplot()

        ax.plot(zdata[:, j], zdata[:, j], c="grey", label="Data")
        ax.scatter(zdata[:, j], zfit[:, j], s=3, c="b", marker=".", label="Predictions")

        ax.set_xlabel(zlabels[j])
        ax.set_ylabel("Model Output")
        ax.set_title("Parity Plot")
        ax.legend()

        if show is True:
            plt.show()
        if filename is not None:
            pdf.savefig(fig)

        fig_list.append(fig)

    if filename is not None:  # place outside loop to avoid closing/reopening
        pdf.close()

    return fig_list


def surrogate_residual(
    surrogate, dataframe, filename=None, relative_error=False, show=True
):
    input_data = dataframe[surrogate.input_labels()]
    output_data = dataframe[surrogate.output_labels()]
    output_surrogate = surrogate.evaluate_surrogate(input_data)
    residual = np.abs(output_data - output_surrogate)
    if relative_error is True:
        residual = np.divide(residual, np.maximum(output_data, 1.0))
    return _residual(
        xdata=input_data.values,
        residual=residual.values,
        xlabels=surrogate.input_labels(),
        elabels=surrogate.output_labels(),
        show=show,
        filename=filename,
    )


def _residual(xdata, residual, xlabels=None, elabels=None, show=True, filename=None):
    """
    Plots model residual against data output.

    Required: input variable data, model error data

    Optional: input and error labels (will be generated as needed), filename
    to save plots to PDF

    Generally, x (input) variables are indexed by 'i', and z (output) variables
    are indexed by 'j'.
    """
    numins = np.shape(xdata)[1]  # number of input variables
    numouts = np.shape(residual)[1]  # number of output variables

    # check and add labels if missing

    if xlabels is None:
        xlabels = []
        for j in range(numins):
            xlabels.append("x" + str(j + 1))
    if elabels is None:
        for i in range(numouts):
            elabels.append("z" + str(j + 1) + " Error")

    if filename is not None:
        if filename[-4:] != ".pdf":  # checking if extension is present
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Filename does not end in .pdf, please amend "
                "filename with the proper extension."
            )
        pdf = PdfPages(filename)

    fig_list = []

    for i in range(numins):
        for j in range(numouts):  # loop over all outputs, zj
            fig = plt.figure()
            ax = fig.add_subplot()
            ax.scatter(xdata[:, i], residual[:, j], marker=".", label=elabels[j])
            ax.set_xlabel(xlabels[i])
            ax.set_ylabel(elabels[j])
            ax.set_title("Residual Plot")
            ax.legend()

            if show is True:
                plt.show()
            if filename is not None:
                pdf.savefig(fig)

            fig_list.append(fig)

    if filename is not None:  # place outside loop to avoid closing/reopening
        pdf.close()

    return fig_list



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for general hierarchical initialization routines
"""
import pytest
import re
import types

from pyomo.environ import Block, ConcreteModel, Constraint, value, Var

from idaes.core.initialization.general_hierarchical import (
    SingleControlVolumeUnitInitializer,
)

from idaes.core.util.exceptions import InitializationError


__author__ = "Andrew Lee"


# TODO: Unit testing of methods
@pytest.mark.unit
def test_nonstandard_model():
    m = ConcreteModel()
    m.initialization_order = [m]

    initializer = SingleControlVolumeUnitInitializer()

    with pytest.raises(
        TypeError,
        match="Model unknown does not appear to be a standard form unit model. "
        "Please use an Initializer specific to the model being initialized.",
    ):
        initializer.initialize(m)


class DummyInit:
    def plugin_prepare(self, plugin, output="default"):
        plugin._test = output

    def plugin_initialize(self, model, **kwargs):
        model._plugin_initialized = True

    def plugin_finalize(self, plugin, output="default"):
        plugin._test2 = output

    def initialize(self, model, **kwargs):
        model._initialized = True


@pytest.mark.unit
def test_init_props_0D_no_copy():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var()

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var()
    m.control_volume.properties_out.v2 = Var()

    def estimate_outlet_state(block, *args, **kwargs):
        block._estimated = True

    m.control_volume.estimate_outlet_state = types.MethodType(
        estimate_outlet_state, m.control_volume
    )

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer._init_props_0D(m.control_volume, False)

    assert m.control_volume.properties_in._initialized
    assert m.control_volume.properties_out._initialized
    assert m.control_volume._estimated


@pytest.mark.unit
def test_init_props_0D_copy():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer._init_props_0D(m.control_volume, True)

    assert m.control_volume.properties_in._initialized
    assert not hasattr(m.control_volume.properties_out, "_initialized")
    assert value(m.control_volume.properties_out.v1) == 12
    assert value(m.control_volume.properties_out.v2) == 7


@pytest.mark.unit
def test_init_props_1D():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.properties = Block()

    initializer = SingleControlVolumeUnitInitializer()

    def estimate_states(block, *args, **kwargs):
        block._estimated = True

    m.control_volume.estimate_states = types.MethodType(
        estimate_states, m.control_volume
    )

    initializer.add_submodel_initializer(m.control_volume.properties, DummyInit())
    initializer._init_props_1D(m.control_volume)

    assert m.control_volume.properties._initialized
    assert m.control_volume._estimated


@pytest.mark.unit
def test_init_rxns():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.reactions = Block()

    # Add a constraint that to the reactions that used a var from outside the block
    m.control_volume.var = Var()
    m.control_volume.reactions.cons = Constraint(expr=m.control_volume.var == 10)

    initializer = SingleControlVolumeUnitInitializer()
    log = initializer.get_logger(m)

    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())
    initializer._init_rxns(m.control_volume)

    assert m.control_volume.reactions._initialized
    assert not m.control_volume.var.fixed


@pytest.mark.unit
def test_initialize_control_volume_no_copy():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    m.control_volume.reactions = Block()
    m.control_volume.reactions.v1 = Var()

    m.control_volume.c1 = Constraint(
        expr=m.control_volume.properties_in.v1 == m.control_volume.properties_out.v1
    )
    m.control_volume.c2 = Constraint(
        expr=m.control_volume.reactions.v1 == 0.5 * m.control_volume.properties_out.v1
    )

    m.control_volume.properties_in.v1.fix()

    def estimate_outlet_state(block, *args, **kwargs):
        block._estimated = True

    m.control_volume.estimate_outlet_state = types.MethodType(
        estimate_outlet_state, m.control_volume
    )

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())

    initializer.initialize_control_volume(m.control_volume, False)

    assert value(m.control_volume.properties_in.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v1) == pytest.approx(6, rel=1e-5)
    assert value(m.control_volume.properties_out.v2) == pytest.approx(7, rel=1e-5)

    assert m.control_volume.properties_in._initialized
    assert m.control_volume.properties_out._initialized
    assert m.control_volume.reactions._initialized
    assert m.control_volume._estimated


@pytest.mark.unit
def test_initialize_control_volume_copy():
    m = ConcreteModel()
    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    m.control_volume.reactions = Block()
    m.control_volume.reactions.v1 = Var()

    m.control_volume.c1 = Constraint(
        expr=m.control_volume.properties_in.v1 == m.control_volume.properties_out.v1
    )
    m.control_volume.c2 = Constraint(
        expr=m.control_volume.reactions.v1 == 0.5 * m.control_volume.properties_out.v1
    )

    m.control_volume.properties_in.v1.fix()

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())

    initializer.initialize_control_volume(m.control_volume, True)

    assert value(m.control_volume.properties_in.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v2) == pytest.approx(7, rel=1e-5)

    assert m.control_volume.properties_in._initialized
    assert not hasattr(m.control_volume.properties_out, "_initialized")
    assert m.control_volume.reactions._initialized


@pytest.mark.unit
def test_initialize_submodels_no_order():
    m = ConcreteModel()
    m.initialization_order = []

    initializer = SingleControlVolumeUnitInitializer()

    with pytest.raises(
        InitializationError,
        match=re.escape(
            "Main model (unknown) was not initialized (no results returned). "
            "This is likely due to an error in the model.initialization_order."
        ),
    ):
        initializer.initialize_submodels(m, {}, {}, copy_inlet_state=False)


@pytest.mark.unit
def test_initialize_submodels_no_plugins_no_copy():
    m = ConcreteModel()
    m.initialization_order = [m]

    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    m.control_volume.reactions = Block()
    m.control_volume.reactions.v1 = Var()

    # plug-in block, but not in initialization order so should be skipped
    m.plugin = Block()

    m.control_volume.c1 = Constraint(
        expr=m.control_volume.properties_in.v1 == m.control_volume.properties_out.v1
    )
    m.control_volume.c2 = Constraint(
        expr=m.control_volume.reactions.v1 == 0.5 * m.control_volume.properties_out.v1
    )

    m.control_volume.properties_in.v1.fix()

    def estimate_outlet_state(block, *args, **kwargs):
        block._estimated = True

    m.control_volume.estimate_outlet_state = types.MethodType(
        estimate_outlet_state, m.control_volume
    )

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())
    initializer.add_submodel_initializer(m.plugin, DummyInit())

    initializer.initialize_submodels(
        m, {m.plugin: {}}, {m.plugin: DummyInit()}, copy_inlet_state=False
    )

    assert value(m.control_volume.properties_in.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v2) == pytest.approx(7, rel=1e-5)
    assert value(m.control_volume.reactions.v1) == pytest.approx(6, rel=1e-5)

    assert m.control_volume.properties_in._initialized
    assert m.control_volume.properties_out._initialized
    assert m.control_volume.reactions._initialized
    assert m.control_volume._estimated
    assert not hasattr(m.plugin, "_initialized")
    assert not hasattr(m.plugin, "_plugin_initialized")


@pytest.mark.unit
def test_initialize_submodels_no_plugins_copy():
    m = ConcreteModel()
    m.initialization_order = [m]

    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    m.control_volume.reactions = Block()
    m.control_volume.reactions.v1 = Var()

    # plug-in block, but not in initialization order so should be skipped
    m.plugin = Block()

    m.control_volume.c1 = Constraint(
        expr=m.control_volume.properties_in.v1 == m.control_volume.properties_out.v1
    )
    m.control_volume.c2 = Constraint(
        expr=m.control_volume.reactions.v1 == 0.5 * m.control_volume.properties_out.v1
    )

    m.control_volume.properties_in.v1.fix()

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())
    initializer.add_submodel_initializer(m.plugin, DummyInit())

    initializer.initialize_submodels(
        m, {m.plugin: {}}, {m.plugin: DummyInit()}, copy_inlet_state=True
    )

    assert value(m.control_volume.properties_in.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v2) == pytest.approx(7, rel=1e-5)
    assert value(m.control_volume.reactions.v1) == pytest.approx(6, rel=1e-5)

    assert m.control_volume.properties_in._initialized
    assert not hasattr(m.control_volume.properties_out, "_initialized")
    assert m.control_volume.reactions._initialized
    assert not hasattr(m.plugin, "_initialized")
    assert not hasattr(m.plugin, "_plugin_initialized")


@pytest.mark.unit
def test_initialize_submodels_w_plugins_copy():
    m = ConcreteModel()
    m.initialization_order = [m]

    m.control_volume = Block()
    m.control_volume.properties_in = Block()
    m.control_volume.properties_in.v1 = Var(initialize=12)

    m.control_volume.properties_out = Block()
    m.control_volume.properties_out.v1 = Var(initialize=6)
    m.control_volume.properties_out.v2 = Var(initialize=7)

    m.control_volume.reactions = Block()
    m.control_volume.reactions.v1 = Var()

    m.plugin = Block()
    m.initialization_order.append(m.plugin)

    m.control_volume.c1 = Constraint(
        expr=m.control_volume.properties_in.v1 == m.control_volume.properties_out.v1
    )
    m.control_volume.c2 = Constraint(
        expr=m.control_volume.reactions.v1 == 0.5 * m.control_volume.properties_out.v1
    )

    m.control_volume.properties_in.v1.fix()

    initializer = SingleControlVolumeUnitInitializer()

    initializer.add_submodel_initializer(m.control_volume.properties_in, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.properties_out, DummyInit())
    initializer.add_submodel_initializer(m.control_volume.reactions, DummyInit())

    initializer.initialize_submodels(
        m, {m.plugin: {}}, {m.plugin: DummyInit()}, copy_inlet_state=True
    )

    assert value(m.control_volume.properties_in.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v1) == pytest.approx(12, rel=1e-5)
    assert value(m.control_volume.properties_out.v2) == pytest.approx(7, rel=1e-5)
    assert value(m.control_volume.reactions.v1) == pytest.approx(6, rel=1e-5)

    assert m.control_volume.properties_in._initialized
    assert not hasattr(m.control_volume.properties_out, "_initialized")
    assert m.control_volume.reactions._initialized
    assert not hasattr(m.plugin, "_initialized")
    assert m.plugin._plugin_initialized



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for setting up FcPh as the state variables in a generic property
package
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from types import MethodType
from pyomo.environ import (
    Constraint,
    Expression,
    NonNegativeReals,
    Var,
    units as pyunits,
)

from idaes.core import MaterialFlowBasis, MaterialBalanceType, EnergyBalanceType

from idaes.models.properties.modular_properties.state_definitions.FTPx import (
    state_initialization,
)
from idaes.models.properties.modular_properties.base.utility import (
    get_bounds_from_config,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from .electrolyte_states import define_electrolyte_state, calculate_electrolyte_scaling

# Set up logger
_log = idaeslog.getLogger(__name__)


# TODO : Need a way to get a guess for T during initialization
def set_metadata(b):
    # Need to update metadata so that enth_mol is recorded as being part of the
    # state variables, and to ensure that getattr does not try to build it
    # using the default method.
    b.get_metadata().properties["enth_mol"].set_method(None)


def define_state(b):
    # FTPx formulation always requires a flash, so set flag to True
    # TODO: should have some checking to make sure developers implement this properly
    b.always_flash = True

    # Check that only necessary state_bounds are defined
    expected_keys = ["flow_mol_comp", "enth_mol", "temperature", "pressure"]
    if (
        b.params.config.state_bounds is not None
        and any(b.params.config.state_bounds.keys()) not in expected_keys
    ):
        for k in b.params.config.state_bounds.keys():
            if k not in expected_keys:
                raise ConfigurationError(
                    "{} - found unexpected state_bounds key {}. Please ensure "
                    "bounds are provided only for expected state variables "
                    "and that you have typed the variable names correctly.".format(
                        b.name, k
                    )
                )

    units = b.params.get_metadata().derived_units
    # Get bounds and initial values from config args
    f_bounds, f_init = get_bounds_from_config(b, "flow_mol_comp", units.FLOW_MOLE)
    h_bounds, h_init = get_bounds_from_config(b, "enth_mol", units.ENERGY_MOLE)
    p_bounds, p_init = get_bounds_from_config(b, "pressure", units.PRESSURE)
    t_bounds, t_init = get_bounds_from_config(b, "temperature", units.TEMPERATURE)

    # Add state variables
    b.flow_mol_comp = Var(
        b.component_list,
        initialize=f_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc="Component molar flowrate",
        units=units.FLOW_MOLE,
    )
    b.pressure = Var(
        initialize=p_init,
        domain=NonNegativeReals,
        bounds=p_bounds,
        doc="State pressure",
        units=units.PRESSURE,
    )
    b.enth_mol = Var(
        initialize=h_init,
        domain=NonNegativeReals,
        bounds=h_bounds,
        doc="Mixture molar specific enthalpy",
        units=units.ENERGY_MOLE,
    )

    # Add supporting variables
    b.flow_mol = Expression(
        expr=sum(b.flow_mol_comp[j] for j in b.component_list),
        doc="Total molar flowrate",
    )

    if f_init is None:
        fp_init = None
    else:
        fp_init = f_init / len(b.phase_list)

    b.flow_mol_phase = Var(
        b.phase_list,
        initialize=fp_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc="Phase molar flow rates",
        units=units.FLOW_MOLE,
    )

    b.temperature = Var(
        initialize=t_init,
        domain=NonNegativeReals,
        bounds=t_bounds,
        doc="Temperature",
        units=units.TEMPERATURE,
    )

    b.mole_frac_comp = Var(
        b.component_list,
        bounds=(1e-20, 1.001),
        initialize=1 / len(b.component_list),
        doc="Mixture mole fractions",
        units=pyunits.dimensionless,
    )

    b.mole_frac_phase_comp = Var(
        b.phase_component_set,
        initialize=1 / len(b.component_list),
        bounds=(1e-20, 1.001),
        doc="Phase mole fractions",
        units=pyunits.dimensionless,
    )

    def Fpc_expr(b, p, j):
        return b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, j]

    b.flow_mol_phase_comp = Expression(
        b.phase_component_set, rule=Fpc_expr, doc="Phase-component molar flowrates"
    )

    b.phase_frac = Var(
        b.phase_list,
        initialize=1.0 / len(b.phase_list),
        bounds=(0, None),
        doc="Phase fractions",
        units=pyunits.dimensionless,
    )

    # Add electrolyte state vars if required
    # This must occur before adding the enthalpy constraint, as it needs true
    # species mole fractions
    if b.params._electrolyte:
        define_electrolyte_state(b)

    # Add supporting constraints
    def rule_mole_frac_comp(b, j):
        if len(b.component_list) > 1:
            return b.flow_mol_comp[j] == b.mole_frac_comp[j] * sum(
                b.flow_mol_comp[k] for k in b.component_list
            )
        else:
            return b.mole_frac_comp[j] == 1.0

    b.mole_frac_comp_eq = Constraint(b.component_list, rule=rule_mole_frac_comp)

    if len(b.phase_list) == 1:

        def rule_total_mass_balance(b):
            return b.flow_mol_phase[b.phase_list.first()] == b.flow_mol

        b.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return (
                b.mole_frac_comp[i] == b.mole_frac_phase_comp[b.phase_list.first(), i]
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_phase_frac(b, p):
            return b.phase_frac[p] == 1.0

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    elif len(b.phase_list) == 2:
        # For two phase, use Rachford-Rice formulation
        def rule_total_mass_balance(b):
            return sum(b.flow_mol_phase[p] for p in b.phase_list) == b.flow_mol

        b.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.flow_mol_comp[i] == sum(
                b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, i]
                for p in b.phase_list
                if (p, i) in b.phase_component_set
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_mole_frac(b):
            return (
                sum(
                    b.mole_frac_phase_comp[b.phase_list.first(), i]
                    for i in b.component_list
                    if (b.phase_list.first(), i) in b.phase_component_set
                )
                - sum(
                    b.mole_frac_phase_comp[b.phase_list.last(), i]
                    for i in b.component_list
                    if (b.phase_list.last(), i) in b.phase_component_set
                )
                == 0
            )

        b.sum_mole_frac = Constraint(rule=rule_mole_frac)

        def rule_phase_frac(b, p):
            return b.phase_frac[p] * b.flow_mol == b.flow_mol_phase[p]

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    else:
        # Otherwise use a general formulation
        def rule_comp_mass_balance(b, i):
            return b.flow_mol_comp[i] == sum(
                b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, i]
                for p in b.phase_list
                if (p, i) in b.phase_component_set
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_mole_frac(b, p):
            return (
                sum(
                    b.mole_frac_phase_comp[p, i]
                    for i in b.component_list
                    if (p, i) in b.phase_component_set
                )
                == 1
            )

        b.sum_mole_frac = Constraint(b.phase_list, rule=rule_mole_frac)

        def rule_phase_frac(b, p):
            return b.phase_frac[p] * b.flow_mol == b.flow_mol_phase[p]

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    # -------------------------------------------------------------------------
    # General Methods
    def get_material_flow_terms_FcPh(b, p, j):
        """Create material flow terms for control volume."""
        return b.flow_mol_phase_comp[p, j]

    b.get_material_flow_terms = MethodType(get_material_flow_terms_FcPh, b)

    def get_enthalpy_flow_terms_FcPh(b, p):
        """Create enthalpy flow terms."""
        # enth_mol_phase probably does not exist when this is created
        # Use try/except to build flow term if not present
        try:
            eflow = b._enthalpy_flow_term
        except AttributeError:

            def rule_eflow(b, p):
                return b.flow_mol_phase[p] * b.enth_mol_phase[p]

            eflow = b._enthalpy_flow_term = Expression(b.phase_list, rule=rule_eflow)
        return eflow[p]

    b.get_enthalpy_flow_terms = MethodType(get_enthalpy_flow_terms_FcPh, b)

    def get_material_density_terms_FcPh(b, p, j):
        """Create material density terms."""
        # dens_mol_phase probably does not exist when this is created
        # Use try/except to build term if not present
        try:
            mdens = b._material_density_term
        except AttributeError:

            def rule_mdens(b, p, j):
                return b.dens_mol_phase[p] * b.mole_frac_phase_comp[p, j]

            mdens = b._material_density_term = Expression(
                b.phase_component_set, rule=rule_mdens
            )
        return mdens[p, j]

    b.get_material_density_terms = MethodType(get_material_density_terms_FcPh, b)

    def get_energy_density_terms_FcPh(b, p):
        """Create energy density terms."""
        # Density and energy terms probably do not exist when this is created
        # Use try/except to build term if not present
        try:
            edens = b._energy_density_term
        except AttributeError:

            def rule_edens(b, p):
                return b.dens_mol_phase[p] * b.energy_internal_mol_phase[p]

            edens = b._energy_density_term = Expression(b.phase_list, rule=rule_edens)
        return edens[p]

    b.get_energy_density_terms = MethodType(get_energy_density_terms_FcPh, b)

    def default_material_balance_type_FcPh():
        return MaterialBalanceType.componentTotal

    b.default_material_balance_type = default_material_balance_type_FcPh

    def default_energy_balance_type_FcPh():
        return EnergyBalanceType.enthalpyTotal

    b.default_energy_balance_type = default_energy_balance_type_FcPh

    def get_material_flow_basis_FcPh():
        return MaterialFlowBasis.molar

    b.get_material_flow_basis = get_material_flow_basis_FcPh

    def define_state_vars_FcPh(b):
        """Define state vars."""
        return {
            "flow_mol_comp": b.flow_mol_comp,
            "enth_mol": b.enth_mol,
            "pressure": b.pressure,
        }

    b.define_state_vars = MethodType(define_state_vars_FcPh, b)

    def define_display_vars_FcPh(b):
        """Define display vars."""
        return {
            "Molar Flowrate": b.flow_mol_comp,
            "Molar Enthalpy": b.enth_mol,
            "Pressure": b.pressure,
        }

    b.define_display_vars = MethodType(define_display_vars_FcPh, b)


def define_default_scaling_factors(b):
    """
    Method to set default scaling factors for the property package. Scaling
    factors are based on the default initial value for each variable provided
    in the state_bounds config argument.
    """
    # Get bounds and initial values from config args
    units = b.get_metadata().derived_units
    state_bounds = b.config.state_bounds

    if state_bounds is None:
        return

    try:
        f_bounds = state_bounds["flow_mol_comp"]
        if len(f_bounds) == 4:
            f_init = pyunits.convert_value(
                f_bounds[1], from_units=f_bounds[3], to_units=units.FLOW_MOLE
            )
        else:
            f_init = f_bounds[1]
    except KeyError:
        f_init = 1

    try:
        p_bounds = state_bounds["pressure"]
        if len(p_bounds) == 4:
            p_init = pyunits.convert_value(
                p_bounds[1], from_units=p_bounds[3], to_units=units.PRESSURE
            )
        else:
            p_init = p_bounds[1]
    except KeyError:
        p_init = 1

    try:
        h_bounds = state_bounds["enth_mol"]
        if len(h_bounds) == 4:
            h_init = pyunits.convert_value(
                h_bounds[1], from_units=h_bounds[3], to_units=units.ENERGY_MOLE
            )
        else:
            h_init = h_bounds[1]
    except KeyError:
        h_init = 1

    try:
        t_bounds = state_bounds["temperature"]
        if len(t_bounds) == 4:
            t_init = pyunits.convert_value(
                t_bounds[1], from_units=t_bounds[3], to_units=units.TEMPERATURE
            )
        else:
            t_init = t_bounds[1]
    except KeyError:
        t_init = 1

    # Set default scaling factors
    b.set_default_scaling("flow_mol", 1 / f_init)
    b.set_default_scaling("flow_mol_phase", 1 / f_init)
    b.set_default_scaling("flow_mol_comp", 1 / f_init)
    b.set_default_scaling("flow_mol_phase_comp", 1 / f_init)
    b.set_default_scaling("pressure", 1 / p_init)
    b.set_default_scaling("temperature", 1 / t_init)
    b.set_default_scaling("enth_mol", 1 / h_init)


def calculate_scaling_factors(b):
    sf_flow = iscale.get_scaling_factor(b.flow_mol, default=1, warning=True)
    sf_h = iscale.get_scaling_factor(b.enth_mol, default=1, warning=True)
    sf_mf = {}
    for i, v in b.mole_frac_phase_comp.items():
        sf_mf[i] = iscale.get_scaling_factor(v, default=1e3, warning=True)

    for j in b.component_list:
        sf_j = iscale.get_scaling_factor(b.mole_frac_comp[j], default=1e3, warning=True)
        iscale.constraint_scaling_transform(
            b.mole_frac_comp_eq[j], sf_j, overwrite=False
        )

    iscale.constraint_scaling_transform(b.enth_mol_eqn, sf_h, overwrite=False)

    if len(b.phase_list) == 1:
        iscale.constraint_scaling_transform(
            b.total_flow_balance, sf_flow, overwrite=False
        )

        for j in b.component_list:
            sf_j = iscale.get_scaling_factor(
                b.mole_frac_comp[j], default=1e3, warning=True
            )
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_j, overwrite=False
            )

        # b.phase_fraction_constraint is well scaled

    elif len(b.phase_list) == 2:
        iscale.constraint_scaling_transform(
            b.total_flow_balance, sf_flow, overwrite=False
        )

        for j in b.component_list:
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_flow, overwrite=False
            )

        iscale.constraint_scaling_transform(
            b.sum_mole_frac, min(sf_mf.values()), overwrite=False
        )

        for p in b.phase_list:
            sf_p = iscale.get_scaling_factor(b.phase_frac[p], default=1, warning=True)
            iscale.constraint_scaling_transform(
                b.phase_fraction_constraint[p], sf_p * sf_flow, overwrite=False
            )

    else:
        for j in b.component_list:
            sf_fc = iscale.get_scaling_factor(
                b.flow_mol_comp[j], default=1, warning=True
            )
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_fc, overwrite=False
            )

        for p in b.phase_list:
            sf_fp = iscale.get_scaling_factor(
                b.flow_mol_phase[p], default=1, warning=True
            )
            iscale.constraint_scaling_transform(
                b.sum_mole_frac[p], min(sf_mf[p, :].values()), overwrite=False
            )
            iscale.constraint_scaling_transform(
                b.phase_fraction_constraint[p], sf_fp, overwrite=False
            )

    if b.params._electrolyte:
        calculate_electrolyte_scaling(b)


# Inherit state_initialization from FTPX form, as the process is the same


do_not_initialize = []


class FcPh(object):
    """Component flow, pressure enthalpy state"""

    set_metadata = set_metadata
    define_state = define_state
    state_initialization = state_initialization
    do_not_initialize = do_not_initialize
    define_default_scaling_factors = define_default_scaling_factors
    calculate_scaling_factors = calculate_scaling_factors



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Tests that helmholtz specific models match generic models"""
import pytest

import pyomo.environ as pyo
import idaes.core
import idaes.models.unit_models as cmodels
import idaes.models_extra.power_generation.unit_models.helm as hmodels
from idaes.models.properties.general_helmholtz import helmholtz_available
from idaes.models.properties import iapws95


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_pump():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Pump(property_package=m.fs.properties)
    m.fs.unit2 = hmodels.HelmPump(property_package=m.fs.properties)
    # set inputs

    Fin = 1e4  # mol/s
    hin = 4000  # J/mol
    Pin = 101325  # Pa
    Pout = 2 * Pin  # Pa
    eff = 0.7
    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_pump.fix(eff)
    m.fs.unit2.efficiency_pump.fix(eff)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7
    )
    assert pyo.value(
        m.fs.unit1.control_volume.properties_out[0].temperature
    ) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature), rel=1e-7
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_turbine():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Turbine(property_package=m.fs.properties)
    m.fs.unit2 = hmodels.HelmIsentropicTurbine(property_package=m.fs.properties)

    # set inputs
    Fin = 1000  # mol/s
    Tin = 500  # K
    Pin = 1e6  # Pa
    Pout = 7e5  # Pa
    hin = pyo.value(iapws95.htpx(Tin * pyo.units.K, Pin * pyo.units.Pa))  # J/mol

    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_isentropic.fix(0.9)
    m.fs.unit2.efficiency_isentropic.fix(0.9)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(
        m.fs.unit1.control_volume.properties_out[0].temperature
    ) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature), rel=1e-7
    )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_compressor():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Compressor(property_package=m.fs.properties)
    m.fs.unit2 = hmodels.HelmIsentropicCompressor(property_package=m.fs.properties)

    # set inputs
    Fin = 1000  # mol/s
    Tin = 500  # K
    Pin = 2222782.4  # Pa
    Pout = 2.7 * Pin  # Pa
    hin = pyo.value(iapws95.htpx(Tin * pyo.units.K, Pin * pyo.units.Pa))  # J/mol
    eff = 0.324

    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_isentropic.fix(eff)
    m.fs.unit2.efficiency_isentropic.fix(eff)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(
        m.fs.unit1.control_volume.properties_out[0].temperature
    ) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature), rel=1e-7
    )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_compressor_pump_compare():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = hmodels.HelmPump(property_package=m.fs.properties)
    m.fs.unit2 = hmodels.HelmIsentropicCompressor(property_package=m.fs.properties)

    Fin = 1e4  # mol/s
    hin = 4000  # J/mol
    Pin = 101325  # Pa
    Pout = 2 * Pin  # Pa
    eff = 0.7
    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_pump.fix(eff)

    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit2.efficiency_isentropic.fix(eff)

    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    # The pump calculations are a bit more approximate assuming incompressible
    # fluid entropy is independent of pressure, so the results here should be
    # close, but not exactly the same.
    assert pyo.value(
        m.fs.unit1.control_volume.properties_out[0].temperature
    ) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature), rel=1e-3
    )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-3
    )
